import uuid

from migration.connector.source.enum import Column

from migration.connector.destination.base import Destination

from migration.connector.source import Source

NUMBER_TYPE = ['BIGINT', 'DECIMAL', 'DOUBLE', 'FLOAT', 'INT', 'SMALLINT', 'TINYINT']
LEFT_BRACKET = '('


def is_number_type(self, column: Column, type_mapping: dict) -> bool:
    if LEFT_BRACKET in column.type:
        column_type = column.type.split(LEFT_BRACKET)[0]
        return type_mapping.get(column_type, column_type) in NUMBER_TYPE

    return type_mapping.get(column.type, column.type) in NUMBER_TYPE


def create_query_result_table(source: Source, destination: Destination, query: str) -> str:
    temp_db = 'validation_query_result_db'
    temp_table = f'validation_query_result_table_{uuid.UUID}'
    source.execute_sql(f"create database if not exists {temp_db}")
    source.execute_sql(f"drop table if exists {temp_db}.{temp_table}")
    source.execute_sql(f"create table {temp_db}.{temp_table} as {query}")
    destination.execute_sql(f"create database if not exists {temp_db}")
    destination.execute_sql(f"drop table if exists {temp_db}.{temp_table}")
    destination.execute_sql(f"create table {temp_db}.{temp_table} as {query}")
    return f"{temp_db}.{temp_table}"


def basic_validation(source: Source, destination: Destination, query: str):
    try:
        tbl_name = create_query_result_table(source, destination, query)
        count_sql = f"select count(*) from {tbl_name}"
        source_count_res = source.execute_sql(count_sql)[0]
        result = {'source_count': source_count_res[0]}
        destination_count_res = destination.execute_sql(count_sql)[0]
        result['destination_count'] = destination_count_res[0]
        type_mapping = source.type_mapping()
        table_columns = source.get_table_columns(tbl_name)
        for column in table_columns:
            if is_number_type(column, type_mapping):
                sql = f"select min({column.name}), max({column.name}), avg({column.name}) from {tbl_name}"
                source_result = source.execute_sql(sql)[0]
                result[f'{column.name}_source_min'] = source_result[0]
                result[f'{column.name}_source_max'] = source_result[1]
                result[f'{column.name}_source_avg'] = source_result[2]
                destination_result = destination.execute_sql(sql)[0]
                result[f'{column.name}_destination_min'] = destination_result[0]
                result[f'{column.name}_destination_max'] = destination_result[1]
                result[f'{column.name}_destination_avg'] = destination_result[2]
        return result
    except Exception as e:
        raise Exception(e)


def construct_profiling_sql(tbl_name: str, source: Source, type_mapping):
    table_columns = source.get_table_columns(tbl_name)
    profile_sql = f"with source_data data as (select * from {tbl_name}), \n" \
                  f"column_profiles  as ( \n"
    for index, column in enumerate(table_columns):
        profile_sql += f"select '{column.name}' as column_name, \n" \
                       f"'{column.type}' as column_type, \n" \
                       f"count(*) as row_count, \n" \
                       f"sum(case when {column.name} is null then 0 else 1 end) / count(*) as not_null_proportion,\n" \
                       f"count(distinct {column.name}) / count(*) as distinct_proportion, \n" \
                       f"count(distinct {column.name}) as distinct_count, \n" \
                       f"count(distinct {column.name}) = count(*) as is_unique, \n"
        if is_number_type(column, type_mapping):
            profile_sql += f"min({column.name}) as min_value, \n" \
                           f"max({column.name}) as max_value, \n" \
                           f"avg({column.name}) as avg_value \n" \
                           f"stddev_pop({column.name}) as stddev_pop_value \n" \
                           f"stddev_sample({column.name}) as stddev_sample_value \n"
        else:
            profile_sql += f"null as min_value, \n" \
                           f"null as max_value, \n" \
                           f"null as avg_value \n" \
                           f"null as stddev_pop_value \n" \
                           f"null as stddev_sample_value \n"
        profile_sql += f"from source_data \n"
        if index != len(table_columns) - 1:
            profile_sql += f"union all \n"
    profile_sql += f") \n" \
                   f"select * from column_profiles;"
    return profile_sql


def multidimensional_validation(source: Source, destination: Destination, query: str):
    try:
        tbl_name = create_query_result_table(source, destination, query)
        type_mapping = source.type_mapping()
        profile_sql = construct_profiling_sql(tbl_name, source, type_mapping)
        source_profile_result = source.execute_sql(profile_sql)
        destination_profile_result = destination.execute_sql(profile_sql)
        result = {'source_profile_result': source_profile_result,
                  'destination_profile_result': destination_profile_result}
        return result
    except Exception as e:
        raise Exception(e)


def line_by_line_validation(query: str, source: Source, destination: Destination):
    try:
        tbl_name = create_query_result_table(source, destination, query)
        source_result = source.execute_sql(f"select * from {tbl_name}")
        destination_result = destination.execute_sql(f"select * from {tbl_name}")
        result = {'source_result': source_result,
                  'destination_result': destination_result}
        return result
    except Exception as e:
        raise Exception(e)



from concurrent.futures import ThreadPoolExecutor

from migration.base.exceptions import SchedulerRuntimeError, ProfileConfigError
from migration.connector.source.base import Source
from migration.connector.destination.base import Destination
from migration.scheduler.scheduler import logger
from migration.util import migration_tasks_status


class Transformer:

    def __init__(self, source: Source, destination: Destination, project_name: str, db_list=None,
                 config_table_list=None,
                 external_table_list=None,
                 scheduler_concurrency=1, quit_if_fail=False, thread_concurrency=1):
        self.source = source
        self.destination = destination
        self.pool = ThreadPoolExecutor(10)
        self.db_list = db_list
        self.project_name = project_name
        self.config_table_list = config_table_list
        self.external_table_list = external_table_list
        self.transform_concurrency = scheduler_concurrency
        self.schedules = None
        self.quit_if_fail = quit_if_fail
        self.schedules_tasks_map = {}
        self.thread_concurrency = thread_concurrency
        self.project_id = migration_tasks_status.init_status_table(self.destination, project_name)

    def transform(self):
        self.schedule_migration_tasks()
        for schedule, tasks in self.schedules_tasks_map.items():
            self.pool.submit(self.run_concurrent, schedule, tasks)
        self.pool.shutdown(wait=True)
        self.source.close()
        self.destination.close()
        logger.info("All tasks are finished")

    def run_concurrent(self, scheduler, tasks):
        try:
            scheduler.tasks_num = len(tasks)
            scheduler.run()
            for task in tasks:
                scheduler.add_task(task)
            scheduler.finish()
        except Exception as e:
            logger.error(f"Scheduler {scheduler} failed to when transforming task, error: {e}")
            raise SchedulerRuntimeError(f"Scheduler {scheduler} failed when transforming task, error: {e}")

    def close(self):
        raise NotImplementedError

    def get_transform_db_list(self):
        return self.db_list

    def get_transform_table_list(self):
        if self.external_table_list:
            return self.external_table_list
        else:
            if self.config_table_list:
                return self.config_table_list
            else:
                if not self.db_list:
                    logger.error("database and table list are both not provided in profile.yml")
                    raise ProfileConfigError("database and table list are both not provided in profile.yml")
                return self.db_list

    def get_migration_tasks(self):
        raise NotImplementedError

    def schedule_migration_tasks(self):
        raise NotImplementedError



from datetime import datetime
from migration.base.status import TaskStatus, TaskType


class Task:
    def __init__(self, name: str, task_type: TaskType, project_id: str, *args, **kwargs):
        self.name = name
        self.task_type = task_type
        self.project_id = project_id
        self.id = self._gen_task_id()
        self.args = args
        self.kwargs = kwargs
        self.status = TaskStatus.INIT
        self.start_time = datetime.now()
        self.end_time = ""
        self.retry_times = 1
        self.status_id = None

    def init_task(self, *args, **kwargs):
        pass

    def _gen_task_id(self):
        return f"{self.name}_{self.task_type.value}_{datetime.now().strftime('%Y%m%d%H%M%S%f')}"

    def __str__(self):
        return self.name + '-' + self.task_type.value

    def __repr__(self):
        return f"<Task: {self.name}, Type:{self.task_type}>"

    def run(self):
        raise NotImplementedError

    def is_failed(self):
        return self.status == TaskStatus.FAILED

    def is_cancelled(self):
        return self.status == TaskStatus.CANCELLED

    def is_success(self):
        return self.status == TaskStatus.COMPLETED

    def is_running(self):
        return self.status == TaskStatus.RUNNING

    def is_queued(self):
        return self.status == TaskStatus.QUEUED

    def is_init(self):
        return self.status == TaskStatus.INIT



class Column:
    def __init__(self, name, type, is_null, default_value=None):
        self.name = name
        self.type = type
        self.is_null = is_null
        self.default_value = default_value


class ClusterInfo:
    def __init__(self, bucket_num, cluster_keys):
        self.bucket_num = bucket_num
        self.cluster_keys = cluster_keys



import logging
import os
import random
from datetime import datetime
from time import sleep

import pymysql
import pymysqlpool as pymysql_pool

from migration.connector.destination.base import Destination
from clickzetta.dbapi.connection import Connection as ClickZettaConnection
from clickzetta.client import Client
from migration.base.exceptions import DestinationExecutionError, GrammarRestrictionsError
from migration.base.status import Status
import migration.util.object_storage_util as object_storage_util

logger = logging.getLogger(__name__)

MAX_GET_STATUS_TIMES = 100


class DorisDestination(Destination):
    def __init__(self, config: dict, meta_conf_path=None, storage_conf_path=None):
        super().__init__('Doris', config)
        self.connection = None
        self.connection_params = self.get_connection_params()
        self.meta_conf_path = meta_conf_path
        self.storage_conf_path = storage_conf_path
        pool_config = {'host': self.connection_params['host'], 'port': self.connection_params['port'],
                       'user': self.connection_params['user'], 'password': self.connection_params['password']}
        self.pool = pymysql_pool.ConnectionPool(size=10, maxsize=50, pre_create_num=5, name='doris_pool', **pool_config)

    def get_connection_params(self):
        assert self.config['fe_servers']
        assert self.config['user']
        assert self.config['password']

        return {
            'host': self.config['fe_servers'][0].split(':')[0],
            'port': int(self.config['fe_servers'][0].split(':')[1]),
            'user': self.config['user'],
            'password': self.config['password']
        }

    def connect(self):
        if self.connection is None:
            self.connection = pymysql.connect(**self.connection_params)
            logger.info(f"Connect to Doris {self.connection_params['host']} successfully")

    def create_table(self, table_name, ddl: str):
        self.execute_sql(ddl)
        logger.info(f"Create table {table_name} successfully")

    def create_database(self, db_name):
        self.execute_sql(f"CREATE DATABASE IF NOT EXISTS {db_name}")
        logger.info(f"Create database {db_name} successfully")

    def execute_sql(self, sql, bind_params=None):
        try:
            connection = self.pool.get_connection()
            with connection.cursor() as cur:
                cur.execute(sql, bind_params)
                result = cur.fetchall()
                connection.close()
                return result
        except Exception as e:
            logger.error("migration Error running SQL: {}, error:{}", sql, e)
            raise DestinationExecutionError('migration Error running SQL: {}, error:{}'.format(sql, e))

    def close(self):
        if self.connection is not None:
            self.connection.close()
        self.pool = None

    def load_external_data(self, file_path, schema_name, table_name):
        object_storage = object_storage_util.get_object_storage_config(self.storage_conf_path)
        label_name = f"{schema_name}_{table_name}_{datetime.now().day}_{random.randint(1, 100000000)}"
        load_sql = f"use {schema_name};\n" \
                   f"LOAD LABEL {label_name}\n" \
                   f"(\n" \
                   f"DATA INFILE(\"s3://{object_storage['bucket']}/{file_path}/*\")\n" \
                   f"INTO TABLE {table_name}\n" \
                   f"FORMAT AS \"parquet\"\n" \
                   f")\n" \
                   f"WITH S3\n" \
                   f"(\n" \
                   f"\"AWS_ENDPOINT\" = \"{object_storage['endpoint']}\",\n" \
                   f"\"AWS_ACCESS_KEY\" = \"{object_storage['id']}\",\n" \
                   f"\"AWS_SECRET_KEY\" = \"{object_storage['key']}\",\n" \
                   f"\"AWS_REGION\" = \"{object_storage['region']}\"\n" \
                   f");\n"
        logger.info(f"Start to load data from {file_path} to {schema_name}.{table_name} with sql: {load_sql}")
        self.execute_sql(load_sql)
        current_times = 0
        while current_times < MAX_GET_STATUS_TIMES:
            if current_times == MAX_GET_STATUS_TIMES - 1:
                raise DestinationExecutionError(
                    f"doris load data from {file_path} to {schema_name}.{table_name}"
                    f" failed, get result more than 100 times.")
            check_state_sql = f"show load where Label = '{label_name}'"
            state = self.execute_sql(check_state_sql)[0][2]
            if state == 'FINISHED':
                break
            elif state == 'CANCELLED' or state == 'FAILED':
                logger.error(f"doris load data from {file_path} to {schema_name}.{table_name} failed")
                raise DestinationExecutionError(
                    f"doris loading data from {file_path} to {schema_name}.{table_name} failed")
            else:
                logger.info(f"Waiting for doris S3 data loading, current state is {state}")
            sleep(5)
            current_times += 1

        logger.info(f"Doirs Load data from {file_path} to {schema_name}.{table_name} successfully")

    def gen_destination_ddl(self, db_name, table_name, columns, primary_keys, cluster_info, partition_columns):
        destination_ddl = 'CREATE TABLE IF NOT EXISTS {db}.{table} (\n'.format(db=db_name, table=table_name)
        for index, column in enumerate(columns):
            destination_ddl += '    {name} {type} {is_null} ' \
                .format(name=column.name,
                        type=column.type,
                        is_null='' if column.is_null else 'NOT NULL')
            # if column.default_value:
            #     destination_ddl += "   DEFAULT {default_value}".format(default_value=column.default_value)
            if index != len(columns) - 1:
                destination_ddl += ',\n'
            else:
                destination_ddl += '\n'

        if primary_keys:
            destination_ddl += '    UNIQUE KEY ({primary_keys}),\n'.format(primary_keys=','.join(primary_keys))
        destination_ddl += ') \n'
        if partition_columns:
            destination_ddl += 'PARTITIONED BY ({partitioned_by})\n'.format(partitioned_by=','.join(partition_columns))
        if cluster_info:
            destination_ddl += 'CLUSTERED BY ({clustered_by})\n'.format(
                clustered_by=','.join(cluster_info.cluster_keys))
            destination_ddl += 'SORTED BY ({clustered_by})\n'.format(clustered_by=','.join(cluster_info.cluster_keys))
            if cluster_info.bucket_num:
                destination_ddl += 'INTO {buckets} BUCKETS\n'.format(buckets=cluster_info.bucket_num)
            else:
                destination_ddl += 'INTO 32 BUCKETS\n'
        self.create_database(db_name)
        return destination_ddl
